// file.h
// Revision 7-nov-2003

#ifndef INCLUDE_FILE_H_
#define INCLUDE_FILE_H_

#ifdef __BORLANDC__
#pragma warn -8022
#endif

#include "dim.h"

#include <iostream>
#include <fstream>
#include <string>
#include <vector>

#if 0
// Not required now, popen use private implementation.

//#ifndef _Windows
#ifndef BLASSIC_USE_WINDOWS
#include <stdio.h> // FILE needed
#else
#include <windows.h> // HANDLE needed
#undef max
#undef min
#endif

#endif

#include "blassic.h"

#if 0
// Not required now, socket use private implementation.
#include "socket.h"
#endif

#include "util.h"

class BlFile {
public:
        enum OpenMode
        	{ Input= 1, Output= 2, Append= 6, Random= 8, Binary= 16 };
        struct field_element {
                size_t size;
                std::string name;
                Dimension dim;
                field_element (size_t n,
                	const std::string & str, const Dimension & dim)
                	:
                        size (n), name (str), dim (dim)
                { }
        };
        enum Align { AlignRight, AlignLeft };

        BlFile (OpenMode nmode);
        virtual ~BlFile ();
        virtual bool isfile ()= 0;
	virtual bool eof ();
	virtual void flush ();
	virtual bool istextwindow () { return false; }
	virtual size_t getwidth ();
	virtual void movecharforward ();
	virtual void movecharforward (size_t n);
	virtual void movecharback ();
	virtual void movecharback (size_t n);
	virtual void movecharup ();
	virtual void movecharup (size_t n);
	virtual void movechardown ();
	virtual void movechardown (size_t n);
	virtual void showcursor ();
	virtual void hidecursor ();
	virtual std::string getkey ();
	virtual std::string inkey ();
	virtual void getline (std::string & str, bool endline= true);
	char delimiter () { return cDelimiter; }
	void delimiter (char delim) { cDelimiter= delim; }
	char quote () { return cQuote; }
	void quote (char qu) { cQuote= qu; }
	char escape () { return cEscape; }
	void escape (char esc) { cEscape= esc; }
        friend BlFile & operator << (BlFile & bf, const std::string & str);
        friend BlFile & operator << (BlFile & bf, char c);
        friend BlFile & operator << (BlFile & bf, BlNumber n);
        friend BlFile & operator << (BlFile & bf, BlInteger n);
        friend BlFile & operator << (BlFile & bf, BlLineNumber l);
        friend BlFile & operator << (BlFile & bf, unsigned short n);
	void putspaces (size_t n);
	virtual void tab ();
	virtual void tab (size_t n);
	virtual void endline ();
	virtual void put (size_t pos);
        virtual void get (size_t pos);
        virtual void field_clear ();
	virtual void field (const std::vector <field_element> & elem);
	virtual void field_append (const std::vector <field_element> & elem);
        virtual bool assign (const std::string & name, const Dimension & dim,
                const std::string & value, Align align);
	virtual bool assign_mid (const std::string & name,
		const Dimension & dim,
                const std::string & value,
                size_t inipos, std::string::size_type len);
	virtual std::string read (size_t n);
	virtual void gotoxy (int x, int y);
	virtual void setcolor (int color);
	virtual int getcolor ();
	virtual void setbackground (int color);
	virtual int getbackground ();
	virtual void cls ();
	virtual std::string copychr ();
	virtual int pos ();
	virtual int vpos ();
	virtual void tag ();
	virtual void tagoff ();
	virtual bool istagactive ();
	virtual void inverse (bool active);
	virtual bool getinverse ();
	virtual void bright (bool active);
	virtual bool getbright ();
	virtual void setwidth (size_t w);
	virtual void setmargin (size_t m);
	virtual BlInteger lof ();
private:
        virtual void outstring (const std::string & str);
        virtual void outchar (char c);
        //virtual void outnumber (BlNumber n);
	//virtual void outinteger (BlInteger n);
        //virtual void outlinenumber (BlLineNumber l);

	BlFile (const BlFile &); // Forbidden
	void operator = (const BlFile &); // Forbidden.

        OpenMode mode;
	char cDelimiter, cQuote, cEscape;
protected:
	OpenMode getmode () const { return mode; }
};

class BlFileConsole : public BlFile {
public:
        BlFileConsole (std::istream & nin, std::ostream & nout);
        bool isfile () { return false; }
	virtual bool eof ();
	virtual void flush ();
	virtual size_t getwidth ();
	virtual void movecharforward ();
	virtual void movecharforward (size_t n);
	virtual void movecharback ();
	virtual void movecharback (size_t n);
	virtual void movecharup ();
	virtual void movecharup (size_t n);
	virtual void movechardown ();
	virtual void movechardown (size_t n);
	virtual void showcursor ();
	virtual void hidecursor ();
	virtual std::string getkey ();
	virtual std::string inkey ();
	void getline (std::string & str, bool endline= true);
	std::string read (size_t n);
	void tab ();
	void tab (size_t n);
	void gotoxy (int x, int y);
	virtual void setcolor (int color);
	virtual void setbackground (int color);
	virtual void cls ();
	int pos ();
private:
        void outstring (const std::string & str);
        void outchar (char c);
	//void outnumber (BlNumber n);
	//void outinteger (BlInteger n);

        std::istream & in;
        std::ostream & out;
        bool ttyin, ttyout;
        //#ifndef _Windows
	#ifndef BLASSIC_USE_WINDOWS
        int xpos;
        #endif
};

class BlFileWindow : public BlFile {
public:
	BlFileWindow (BlChannel ch);
	BlFileWindow (BlChannel ch, int x1, int x2, int y1, int y2);
	~BlFileWindow ();
	bool isfile () { return false; }
	void reset (int x1, int x2, int y1, int y2);
	virtual bool eof ();
	virtual void flush ();
	virtual size_t getwidth ();
	virtual void movecharforward ();
	virtual void movecharforward (size_t n);
	virtual void movecharback ();
	virtual void movecharback (size_t n);
	virtual void movecharup ();
	virtual void movecharup (size_t n);
	virtual void movechardown ();
	virtual void movechardown (size_t n);
	virtual void showcursor ();
	virtual void hidecursor ();
	virtual std::string getkey ();
	virtual std::string inkey ();
	virtual bool istextwindow () { return true; }
	void getline (std::string & str, bool endline= true);
	std::string read (size_t n);
	void tab ();
	void tab (size_t n);
	void endline ();
	void gotoxy (int x, int y);
	virtual void setcolor (int color);
	virtual int getcolor ();
	virtual void setbackground (int color);
	virtual int getbackground ();
	virtual void cls ();
	std::string copychr ();
	int pos ();
	int vpos ();
	void tag ();
	void tagoff ();
	bool istagactive ();
	void inverse (bool active);
	bool getinverse ();
	void bright (bool active);
	bool getbright ();
private:
        void outstring (const std::string & str);
        void outchar (char c);

	BlChannel ch;
};

class BlFileOut : public BlFile {
public:
	BlFileOut ();
	BlFileOut (OpenMode mode);
	void flush ();
protected:
	virtual std::ostream & ofs ()= 0;
private:
	void outstring (const std::string & str);
	void outchar (char c);
	//void outnumber (BlNumber n);
	//void outinteger (BlInteger n);
	//void outlinenumber (BlLineNumber l);
};

class BlFileOutString : public BlFileOut {
public:
	BlFileOutString ();
	bool isfile () { return false; }
	std::string str ();
private:
	std::ostream & ofs ();
	std::ostringstream oss;
};

class BlFileOutput : public BlFileOut {
public:
	BlFileOutput (std::ostream & os);
        bool isfile () { return true; }
private:
	std::ostream & ofs ();
	std::ostream & os;
};

class BlFileRegular : public BlFileOut {
public:
        BlFileRegular (const std::string & name, OpenMode mode);
	bool isfile () { return true; }
	void getline (std::string & str, bool endline= true);
	bool eof ();
	void flush ();
	virtual std::string getkey ();
	virtual std::string inkey ();
	std::string read (size_t n);
	BlInteger lof ();
private:
	std::ostream & ofs ();
        std::fstream fs;
};

class BlFileRandom : public BlFile {
public:
	BlFileRandom (const std::string & name, size_t record_len);
	bool isfile () { return true; }
	void put (size_t pos);
	void get (size_t pos);
	void field_clear ();
	void field (const std::vector <field_element> & elem);
	void field_append (const std::vector <field_element> & elem);
        virtual bool assign (const std::string & name, const Dimension & dim,
                const std::string & value, Align align);
	virtual bool assign_mid (const std::string & name,
		const Dimension & dim,
                const std::string & value,
                size_t inipos, std::string::size_type len);
        struct field_chunk {
                std::string name;
                Dimension dim;
                size_t pos;
                size_t size;
		inline void getvar (char * buf) const;
        };
        typedef std::vector <field_chunk> vchunk;
        BlInteger lof ();
private:
	std::fstream fs;
	size_t len;
	size_t len_assigned;
	size_t actual;
	//auto_array buf;
	util::auto_buffer <char> buf;
	vchunk chunk;
};

class BlFilePopen : public BlFile {
public:
        BlFilePopen (const std::string & str, OpenMode mode);
        ~BlFilePopen ();
	bool isfile () { return true; }
	void getline (std::string & str, bool endline= true);
	bool eof ();
	void flush ();
	std::string read (size_t n);
private:
        void outstring (const std::string & str);
        void outchar (char c);

	#if 1
	class Internal;
	Internal * pin;
	#else
        static const size_t bufsize= 4096;
        char buffer [bufsize];
        size_t bufpos, bufread;
        void readbuffer ();
        char getcharfrombuffer ();
        //#ifdef _Windows
	#ifdef BLASSIC_USE_WINDOWS
        HANDLE hpipe;
        #else
	FILE * hpipe;
        #endif
        #endif
};

class BlFileSocket : public BlFile {
public:
	BlFileSocket (const std::string & host, short port);
	~BlFileSocket ();
	bool isfile () { return true; }
	void getline (std::string & str, bool endline= true);
	bool eof ();
	void flush ();
	std::string read (size_t n);
private:
	void outstring (const std::string & str);
	void outchar (char c);

	class Internal;
	Internal * pin;
	//TcpSocketClient socket;
};

class BlFilePrinter : public BlFile {
public:
	BlFilePrinter ();
	~BlFilePrinter ();
	bool isfile () { return false; }
	void flush ();
	size_t getwidth ();
	void tab ();
	void tab (size_t n);
	void endline ();
	void setwidth (size_t w);
	void setmargin (size_t m);
private:
	void outstring (const std::string & str);
	void outchar (char c);
	class Internal;
	Internal * pi;
};


#endif

// Fin de file.h
